#!/usr/bin/env python

import sys

import json
import yaml

from charmhelpers.contrib.openstack.utils import configure_installation_source

from charmhelpers.core.hookenv import (
    Hooks,
    UnregisteredHookError,
    config,
    local_unit,
    log,
    relation_get,
    relation_ids,
    relation_set,
    remote_unit
)

from charmhelpers.core.host import (
    pwgen,
    restart_on_change,
    service_restart
)

from charmhelpers.fetch import (
    apt_install,
    apt_upgrade,
    configure_sources
)

from contrail_configuration_utils import (
    api_port,
    check_url,
    contrail_floating_ip_create,
    contrail_floating_ip_deactivate,
    contrail_floating_ip_delete,
    contrail_floating_ip_use,
    fix_permissions,
    provision_metadata,
    units,
    unprovision_metadata,
    write_contrail_api_config,
    write_contrail_schema_config,
    write_discovery_config,
    write_ifmap_config,
    write_vnc_api_config
)

PACKAGES = [ "ifmap-server", "contrail-config", "contrail-config-openstack",
             "neutron-common", "contrail-utils" ]

hooks = Hooks()
config = config()

def add_contrail_api():
    if not config_get("contrail-api-configured") \
       and config_get("amqp-ready") \
       and config_get("cassandra-ready") \
       and config_get("identity-admin-ready") \
       and config_get("zookeeper-ready"):
        api_p = api_port()
        check_url("http://localhost:" + str(api_p))
        config["contrail-api-configured"] = True
        for rid in relation_ids("contrail-api"):
            relation_set(relation_id=rid, port=api_p)

def add_metadata():
    if not config_get("metadata-provisioned") \
       and config_get("contrail-api-configured") \
       and config_get("neutron-metadata-ready"):
        provision_metadata()
        config["metadata-provisioned"] = True

@hooks.hook("amqp-relation-changed")
def amqp_changed():
    if not relation_get("password"):
        log("Relation not ready")
        return
    amqp_relation()
    config["amqp-ready"] = True
    add_contrail_api()
    add_metadata()

@hooks.hook("amqp-relation-departed")
@hooks.hook("amqp-relation-broken")
def amqp_departed():
    if not units("amqp"):
        remove_metadata()
        remove_contrail_api()
        config["amqp-ready"] = False
    amqp_relation()

@restart_on_change({"/etc/contrail/contrail-api.conf": ["contrail-api"]})
def amqp_relation():
    write_contrail_api_config()

@hooks.hook("amqp-relation-joined")
def amqp_joined():
    relation_set(username="contrail", vhost="contrail")

@hooks.hook("cassandra-relation-changed")
def cassandra_changed():
    if not relation_get("port"):
        log("Relation not ready")
        return
    cassandra_relation()
    config["cassandra-ready"] = True
    add_contrail_api()
    add_metadata()

@hooks.hook("cassandra-relation-departed")
@hooks.hook("cassandra-relation-broken")
def cassandra_departed():
    if not units("cassandra"):
        remove_metadata()
        remove_contrail_api()
        config["cassandra-ready"] = False
    cassandra_relation()

@restart_on_change({"/etc/contrail/contrail-api.conf": ["contrail-api"],
                    "/etc/contrail/contrail-discovery.conf": ["contrail-discovery"],
                    "/etc/contrail/contrail-schema.conf": ["contrail-schema"],
                    "/etc/contrail/discovery.conf": ["contrail-discovery"]})
def cassandra_relation():
    write_contrail_api_config()
    write_contrail_schema_config()
    write_discovery_config()

@hooks.hook("config-changed")
def config_changed():
    configure_floating_ip_pools()

def config_get(key):
    try:
        return config[key]
    except KeyError:
        return None

def configure_floating_ip_pools():
    floating_pools = config.get("floating-ip-pools")
    previous_floating_pools = config_get("floating-ip-pools-prev")
    if floating_pools != previous_floating_pools:
        pools = { (pool["project"],
                   pool["network"],
                   pool["pool-name"]): set(pool["target-projects"])
                  for pool in yaml.safe_load(floating_pools) } \
                if floating_pools else {}
        previous_pools = {}
        if previous_floating_pools:
            for pool in yaml.safe_load(previous_floating_pools):
                projects = pool["target-projects"]
                name = (pool["project"], pool["network"], pool["pool-name"])
                if name in pools:
                    previous_pools[name] = set(projects)
                else:
                    floating_ip_pool_delete(name, projects)
        for name, projects in pools.iteritems():
            if name not in previous_pools:
                floating_ip_pool_create(name, projects)
            else:
                floating_ip_pool_update(name, projects, previous_pools[name])

        config["floating-ip-pools-prev"] = floating_pools

@hooks.hook("contrail-api-relation-joined")
def contrail_api_joined():
    if config_get("contrail-api-configured"):
        relation_set(port=api_port())

@hooks.hook("contrail-ifmap-relation-broken")
@restart_on_change({"/etc/ifmap-server/basicauthusers.properties": ["ifmap-server"]})
def contrail_ifmap_broken():
    write_ifmap_config()

@hooks.hook("contrail-ifmap-relation-departed")
def contrail_ifmap_departed():
    creds = relation_get("creds", local_unit())
    creds = json.loads(creds) if creds else {}
    unit = remote_unit()
    if unit not in creds:
        return
    del creds[unit]
    relation_set(creds=json.dumps(creds))
    write_ifmap_config()
    service_restart("ifmap-server")

@hooks.hook("contrail-ifmap-relation-joined")
def contrail_ifmap_joined():
    creds = relation_get("creds", local_unit())
    creds = json.loads(creds) if creds else {}
    unit = remote_unit()
    if unit in creds:
        return
    creds[unit] = { "username": unit, "password": pwgen(32) }
    relation_set(creds=json.dumps(creds))
    write_ifmap_config()
    service_restart("ifmap-server")

def floating_ip_pool_create(name, projects):
    fq_network = "default-domain:" + ":".join(name[:2])
    contrail_floating_ip_create(fq_network, name[2])

    fq_pool_name = "default-domain:" + ":".join(name)
    for project in projects:
        fq_project = "default-domain:" + project
        contrail_floating_ip_use(fq_project, fq_pool_name)

def floating_ip_pool_delete(name, projects):
    fq_pool_name = "default-domain:" + ":".join(name)
    for project in projects:
        fq_project = "default-domain:" + project
        contrail_floating_ip_deactivate(fq_project, fq_pool_name)

    fq_network = "default-domain:" + ":".join(name[:2])
    contrail_floating_ip_delete(fq_network, name[2])

def floating_ip_pool_update(name, projects, previous_projects):
    fq_pool_name = "default-domain:" + ":".join(name)
    for project in (previous_projects - projects):
        fq_project = "default-domain:" + project
        contrail_floating_ip_deactivate(fq_project, fq_pool_name)
    for project in (projects - previous_projects):
        fq_project = "default-domain:" + project
        contrail_floating_ip_use(fq_project, fq_pool_name)

@hooks.hook("identity-admin-relation-changed")
def identity_admin_changed():
    if not relation_get("service_hostname"):
        log("Relation not ready")
        return
    identity_admin_relation()
    config["identity-admin-ready"] = True
    add_contrail_api()
    add_metadata()

@hooks.hook("identity-admin-relation-departed")
@hooks.hook("identity-admin-relation-broken")
def identity_admin_departed():
    if not units("identity-admin"):
        remove_metadata()
        remove_contrail_api()
        config["identity-admin-ready"] = False
    identity_admin_relation()

@restart_on_change({"/etc/contrail/contrail-api.conf": ["contrail-api"],
                    "/etc/contrail/contrail-schema.conf": ["contrail-schema"]})
def identity_admin_relation():
    write_contrail_api_config()
    write_contrail_schema_config()
    write_vnc_api_config()

@hooks.hook()
def install():
    configure_installation_source(config["openstack-origin"])
    configure_sources(True, "install-sources")
    apt_upgrade(fatal=True, dist=True)
    apt_install(PACKAGES, fatal=True)
    fix_permissions()
    # broken java dependency on contrail 2.0+
    service_restart("ifmap-server")

def main():
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log("Unknown hook {} - skipping.".format(e))

@hooks.hook("neutron-metadata-relation-changed")
def neutron_metadata_changed():
    if not relation_get("shared-secret"):
        log("Relation not ready")
        return
    config["neutron-metadata-ready"] = True
    add_metadata()

@hooks.hook("neutron-metadata-relation-departed")
@hooks.hook("neutron-metadata-relation-broken")
def neutron_metadata_departed():
    if not units("neutron-metadata"):
        remove_metadata()
        config["neutron-metadata-ready"] = False

def remove_contrail_api():
    config["contrail-api-configured"] = False

def remove_metadata():
    if config_get("metadata-provisioned"):
        unprovision_metadata()
        config["metadata-provisioned"] = False

@hooks.hook("upgrade-charm")
def upgrade_charm():
    write_ifmap_config()
    write_contrail_api_config()
    write_contrail_schema_config()
    write_discovery_config()
    write_vnc_api_config()
    for service in [ "ifmap-server", "contrail-api", "contrail-schema",
                     "contrail-discovery" ]:
        service_restart(service)

@hooks.hook("zookeeper-relation-changed")
def zookeeper_changed():
    if not relation_get("port"):
        log("Relation not ready")
        return
    zookeeper_relation()
    config["zookeeper-ready"] = True
    add_contrail_api()
    add_metadata()

@hooks.hook("zookeeper-relation-departed")
@hooks.hook("zookeeper-relation-broken")
def zookeeper_departed():
    if not units("zookeeper"):
        remove_metadata()
        remove_contrail_api()
        config["zookeeper-ready"] = False
    zookeeper_relation()

@restart_on_change({"/etc/contrail/contrail-api.conf": ["contrail-api"],
                    "/etc/contrail/contrail-discovery.conf": ["contrail-discovery"],
                    "/etc/contrail/contrail-schema.conf": ["contrail-schema"],
                    "/etc/contrail/discovery.conf": ["contrail-discovery"]})
def zookeeper_relation():
    write_contrail_api_config()
    write_contrail_schema_config()
    write_discovery_config()

if __name__ == "__main__":
    main()
