import functools
import os
from socket import gethostbyname
from subprocess import (
    CalledProcessError,
    check_call,
    check_output
)
from time import sleep, time

import apt_pkg
from apt_pkg import version_compare
import json
import urllib2

from charmhelpers.core.hookenv import (
    local_unit,
    log,
    related_units,
    relation_get,
    relation_ids,
    relation_type,
    remote_unit,
    unit_get
)

from charmhelpers.core.templating import render

apt_pkg.init()

def dpkg_version(pkg):
    try:
        return check_output(["dpkg-query", "-f", "${Version}\\n", "-W", pkg]).rstrip()
    except CalledProcessError:
        return None

CONTRAIL_VERSION = dpkg_version("contrail-config")

def retry(f=None, timeout=10, delay=2):
    if not f:
        return functools.partial(retry, timeout=timeout, delay=delay)
    @functools.wraps(f)
    def func(*args, **kwargs):
        start = time()
        error = None
        while True:
            try:
                return f(*args, **kwargs)
            except Exception as e:
                error = e
            elapsed = time() - start
            if elapsed >= timeout:
                raise error
            remaining = timeout - elapsed
            if delay <= remaining:
                sleep(delay)
            else:
                sleep(remaining)
                raise error
    return func

def amqp_ctx():
    ctxs = [ { "rabbit_server": gethostbyname(relation_get("hostname", unit, rid)),
               "rabbit_user": "contrail",
               "rabbit_password": password,
               "rabbit_vhost": "contrail" }
             for rid in relation_ids("amqp")
             for unit, password in
             ((unit, relation_get("password", unit, rid)) for unit in related_units(rid))
             if password ]
    return ctxs[0] if ctxs else {}

def api_port():
    return 9100 if version_compare(CONTRAIL_VERSION, "2.0") >= 0 else 8082

def cassandra_ctx():
    return { "cassandra_servers": [ gethostbyname(relation_get("private-address", unit, rid))
                                    + ":" + port
                                    for rid in relation_ids("cassandra")
                                    for unit, port in
                                    ((unit, relation_get("port", unit, rid)) for unit in related_units(rid))
                                    if port ] }

@retry(timeout=120)
def check_url(url):
    try:
        urllib2.urlopen(url)
    except urllib2.HTTPError:
        pass

def contrail_ctx():
    addr = gethostbyname(unit_get("private-address"))
    return { "api_port": api_port(), "ifmap_server": addr, "disc_server": addr }

def contrail_floating_ip_create(network, name):
    user, password, tenant = [ (relation_get("service_username", unit, rid),
                                relation_get("service_password", unit, rid),
                                relation_get("service_tenant_name", unit, rid))
                               for rid in relation_ids("identity-admin")
                               for unit in related_units(rid) ][0]
    log("Creating floating ip pool {} for network {}".format(name, network))
    check_call(["python", "/usr/share/contrail-utils/create_floating_pool.py",
                "--public_vn_name", network,
                "--floating_ip_pool_name", name,
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--admin_user", user,
                "--admin_password", password,
                "--admin_tenant", tenant])

def contrail_floating_ip_deactivate(project, name):
    user, password, tenant = [ (relation_get("service_username", unit, rid),
                                relation_get("service_password", unit, rid),
                                relation_get("service_tenant_name", unit, rid))
                               for rid in relation_ids("identity-admin")
                               for unit in related_units(rid) ][0]
    log("Deactivating floating ip pool {} for project {}".format(name, project))
    check_call(["scripts/deactivate_floating_pool.py",
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--admin_user", user,
                "--admin_password", password,
                "--admin_tenant", tenant,
                project, name])

def contrail_floating_ip_delete(network, name):
    user, password, tenant = [ (relation_get("service_username", unit, rid),
                                relation_get("service_password", unit, rid),
                                relation_get("service_tenant_name", unit, rid))
                               for rid in relation_ids("identity-admin")
                               for unit in related_units(rid) ][0]
    log("Deleting floating ip pool {} for network {}".format(name, network))
    check_call(["scripts/delete_floating_pool.py",
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--admin_user", user,
                "--admin_password", password,
                "--admin_tenant", tenant,
                network, name])

def contrail_floating_ip_use(project, name):
    user, password, tenant = [ (relation_get("service_username", unit, rid),
                                relation_get("service_password", unit, rid),
                                relation_get("service_tenant_name", unit, rid))
                               for rid in relation_ids("identity-admin")
                               for unit in related_units(rid) ][0]
    log("Activating floating ip pool {} for project {}".format(name, project))
    check_call(["python", "/usr/share/contrail-utils/use_floating_pool.py",
                "--project_name", project,
                "--floating_ip_pool_name", name,
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--admin_user", user,
                "--admin_password", password,
                "--admin_tenant", tenant])

def contrail_ifmap_ctx():
    creds = []
    unit = local_unit()
    for rid in relation_ids("contrail-ifmap"):
        cs = relation_get("creds", unit, rid)
        if cs:
            cs = json.loads(cs)
            for c in cs.itervalues():
                creds.append(c)
    return { "ifmap_creds": creds }

def fix_permissions():
    os.chmod("/etc/contrail", 0755)
    os.chown("/etc/contrail", 0, 0)

def identity_admin_ctx():
    ctxs = [ { "auth_host": gethostbyname(hostname),
               "auth_port": relation_get("service_port", unit, rid),
               "admin_user": relation_get("service_username", unit, rid),
               "admin_password": relation_get("service_password", unit, rid),
               "admin_tenant_name": relation_get("service_tenant_name", unit, rid) }
             for rid in relation_ids("identity-admin")
             for unit, hostname in
             ((unit, relation_get("service_hostname", unit, rid)) for unit in related_units(rid))
             if hostname ]
    return ctxs[0] if ctxs else {}

def provision_metadata():
    ip = [ gethostbyname(relation_get("private-address", unit, rid))
           for rid in relation_ids("neutron-metadata")
           for unit in related_units(rid) ][0]
    user, password = [ (relation_get("service_username", unit, rid),
                        relation_get("service_password", unit, rid))
                       for rid in relation_ids("identity-admin")
                       for unit in related_units(rid) ][0]
    log("Provisioning metadata service {}:8775".format(ip))
    check_call(["contrail-provision-linklocal",
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--linklocal_service_name", "metadata",
                "--linklocal_service_ip", "169.254.169.254",
                "--linklocal_service_port", "80",
                "--ipfabric_service_ip", ip,
                "--ipfabric_service_port", "8775",
                "--oper", "add",
                "--admin_user", user,
                "--admin_password", password])

def units(relation):
    return [ unit for rid in relation_ids(relation)
                  for unit in related_units(rid) ]

def unprovision_metadata():
    if not remote_unit():
        return
    relation = relation_type()
    ip = None
    if relation == "neutron-metadata":
        ip = gethostbyname(relation_get("private-address"))
    else:
        ip = [ gethostbyname(relation_get("private-address", unit, rid))
               for rid in relation_ids("neutron-metadata")
               for unit in related_units(rid) ][0]
    user = None
    password = None
    if relation == "identity-admin":
        user = relation_get("service_username")
        password = relation_get("service_password")
    else:
        user, password = [ (relation_get("service_username", unit, rid),
                            relation_get("service_password", unit, rid))
                           for rid in relation_ids("identity-admin")
                           for unit in related_units(rid) ][0]
    log("Unprovisioning metadata service {}:8775".format(ip))
    check_call(["contrail-provision-linklocal",
                "--api_server_ip", "127.0.0.1",
                "--api_server_port", str(api_port()),
                "--linklocal_service_name", "metadata",
                "--linklocal_service_ip", "169.254.169.254",
                "--linklocal_service_port", "80",
                "--ipfabric_service_ip", ip,
                "--ipfabric_service_port", "8775",
                "--oper", "del",
                "--admin_user", user,
                "--admin_password", password])

def write_contrail_api_config():
    ctx = {}
    ctx.update(contrail_ctx())
    ctx.update(cassandra_ctx())
    ctx.update(zookeeper_ctx())
    ctx.update(amqp_ctx())
    ctx.update(identity_admin_ctx())
    render("contrail-api.conf", "/etc/contrail/contrail-api.conf", ctx, "root",
           "contrail", 0440)

def write_contrail_schema_config():
    ctx = {}
    ctx.update(cassandra_ctx())
    ctx.update(zookeeper_ctx())
    ctx.update(identity_admin_ctx())
    ctx.update(contrail_ctx())
    render("contrail-schema.conf", "/etc/contrail/contrail-schema.conf",
           ctx, "root", "contrail", 0440)

def write_discovery_config():
    ctx = {}
    ctx.update(zookeeper_ctx())
    ctx.update(cassandra_ctx())
    target = "/etc/contrail/contrail-discovery.conf" \
             if version_compare(CONTRAIL_VERSION, "1.20~") >= 0 \
             else "/etc/contrail/discovery.conf"
    render("discovery.conf", target, ctx)

def write_ifmap_config():
    ctx = contrail_ifmap_ctx()
    render("basicauthusers.properties",
           "/etc/ifmap-server/basicauthusers.properties", ctx, "root",
           "contrail", 0440)

def write_vnc_api_config():
    ctx = {}
    ctx.update(contrail_ctx())
    ctx.update(identity_admin_ctx())
    render("vnc_api_lib.ini", "/etc/contrail/vnc_api_lib.ini", ctx)

def zookeeper_ctx():
    return { "zk_servers": [ gethostbyname(relation_get("private-address", unit, rid))
                             + ":" + port
                             for rid in relation_ids("zookeeper")
                             for unit, port in
                             ((unit, relation_get("port", unit, rid)) for unit in related_units(rid))
                             if port ] }
